; Mazpa: a CKY parser for 2DCF grammars
; MPradella, v. 20060323 

; Written in PLT Scheme (v. 4): http://www.plt-scheme.org/
; 
; --------------------------------------------------------------------------
;
; Copyright (C) 2006-2008 Matteo Pradella (pradella@elet.polimi.it)
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
;
; --------------------------------------------------------------------------

#lang scheme/base

(provide
  parse!
  set-rules!
  toggle-trace!)

(require 
  (file "mazpa-utils.scm")
  (only-in srfi/1 lset-union)
  srfi/25)


; --- globals ---
;
(define *the-matrix*    #f)
(define *t-rules*       #f)
(define *h-rules*       #f)
(define *v-rules*       #f)
(define *please-trace-parsing* #f)

(define (set-rules! t h v)
  (set! *t-rules* t)
  (set! *h-rules* h)
  (set! *v-rules* v))

(define (toggle-trace!)
  (set! *please-trace-parsing* (not *please-trace-parsing*)))


; --- data ---


(define (make-structure m n)
  (make-array (shape 0 m 0 n 0 m 0 n) '()))

(define (subpict-ref rec)
  (let ((i (vector-ref rec 0))(j (vector-ref rec 1))
	(r (vector-ref rec 2))(s (vector-ref rec 3)))
    (array-ref *the-matrix* (sub1 i) (sub1 j) (sub1 r) (sub1 s))))


(define (subpict-set! what rec)
  (let ((i (vector-ref rec 0))(j (vector-ref rec 1))
	(r (vector-ref rec 2))(s (vector-ref rec 3)))
    (array-set! *the-matrix* (sub1 i) (sub1 j) (sub1 r) (sub1 s) what)))

(define (load-pict! pict)
  (let* ((v-max (length pict))
	 (h-max (length (car pict))))
    (set! *the-matrix* (make-structure v-max h-max))
    (for-range
      1 v-max
      (lambda (v)
	(for-range
	  1 h-max		   
	  (lambda (h)
	    (subpict-set!
	      (cdr (assoc (list-ref (list-ref pict (sub1 v)) (sub1 h)) *t-rules*))
	      `#(,v ,h ,v ,h))))))))

; --- the algorithm ---
; Given a picture p, such that |p| = (m,n):
; Time complexity: O((m.n)^2 (m+n)), 
; Space complexity: Theta((m.n)^2)

(define (v-compute-subrec rec)
  (let ((out '())
	(i (vector-ref rec 0))(j (vector-ref rec 1))
	(r (vector-ref rec 2))(s (vector-ref rec 3)))

    (for-range
      i (- r 1)
      (lambda (k)
	(set! out (cons
		    (cons `#(,i ,j ,k ,s) `#(,(+ 1 k) ,j ,r ,s))
		    out))))
    out))

(define (h-compute-subrec rec)
  (let ((out '())
	(i (vector-ref rec 0))(j (vector-ref rec 1))
	(r (vector-ref rec 2))(s (vector-ref rec 3)))

    (for-range
      j (- s 1)
      (lambda (k)
	(set! out (cons
		    (cons `#(,i ,j ,r ,k) `#(,i ,(+ 1 k) ,r ,s))
		    out))))
    out))

(define (add-nonterminal nonterm rec)
  (when *please-trace-parsing*
    (begin (display nonterm)(newline)))
  (subpict-set!
    (lset-union equal? nonterm (subpict-ref rec))
    rec))

(define (h-step rec)
  (define subr (h-compute-subrec rec))

  (for-each
    (lambda (p)
      (let ((left (subpict-ref (car p)))
	    (right (subpict-ref (cdr p))))
	(for-each
	  (lambda (l)
	    (for-each
	      (lambda (r)
		(let ((nonterm (assoc (cons l r) *h-rules*)))
		  (when nonterm
		    (add-nonterminal (cdr nonterm) rec))))
	      right))
	  left)))
    subr))

(define (v-step rec)
  (define subr (v-compute-subrec rec))

  (for-each
    (lambda (p)
      (let ((left (subpict-ref (car p)))
	    (right (subpict-ref (cdr p))))
	(for-each
	  (lambda (l)
	    (for-each
	      (lambda (r)
		(let ((nonterm (assoc (cons l r) *v-rules*)))
		  (when nonterm
		    (add-nonterminal (cdr nonterm) rec))))
	      right))
	  left)))
    subr))

(define (subpict-loop v-max h-max v-size h-size)
  (for-range
    1 v-max
    (lambda (v)
      (for-range
	1 h-max
	(lambda (h)
	  (let* ((vv (+ v (sub1 v-size)))
		 (hh (+ h (sub1 h-size)))
		 (rec `#(,v ,h ,vv ,hh)))
	    (when (and (<=  vv v-max) 
		       (<=  hh h-max))
	      (begin
		(when *please-trace-parsing*
		  (begin (display rec)(newline)))
		(v-step rec)
		(h-step rec)))))))))

(define (size-loop)
  (let* ((h-max (array-end *the-matrix* 1))
	 (v-max (array-end *the-matrix* 0)))
    (for-range
      1 v-max
      (lambda (v)
	(for-range
	  1 h-max		   
	  (lambda (h)
	    (when (> (* h v) 1)
	      (subpict-loop v-max h-max v h))))))
    (display "Output: ")
    (display (if (member 'S
			 (array-ref *the-matrix* 0 0 (sub1 v-max) (sub1 h-max)))
	       "Yes"
	       "No"))
    (newline)))

; --- MAIN ---
(define (parse! pict)
  (load-pict! pict)
  (size-loop))

