; Mazpa: parsers for 2D grammars
; Utilities
;
; MPradella, v. 20060323 

; Written in PLT Scheme (v. 4): http://www.plt-scheme.org/
; 
; --------------------------------------------------------------------------
;
; Copyright (C) 2006-2008 Matteo Pradella (pradella@elet.polimi.it)
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
;
; --------------------------------------------------------------------------

#lang scheme/base

(provide
  for-range
  for-range-2
  cartprod
  cons-no-void
  )

; --- macros ---


; simple for-range (with optional step)
(define-syntax for-range
  (syntax-rules ()
		((_ from to fun)
		 (do ((i from (+ i 1)))
		   ((> i to))
		   (fun i)))
		((_ from to step fun)
		 (do ((i from (+ i step)))
		   ((> i to))
		   (fun i)))))

; simple 2D for-range
(define-syntax for-range-2
  (syntax-rules ()
    ((_ from1 to1 from2 to2 fun)
     (do ((i from1 (add1 i)))
       ((> i to1))
       (do ((j from2 (add1 j)))
         ((> j to2))
         (fun i j))))))


; cartesian product of list of lists
(define (pcar A B)
  (for*/list ((a (in-list A))
              (b (in-list B)))
             ((if (list? b) cons list) a b)))

(define (cartprod ls)
  (foldl pcar (car ls) (cdr ls)))


;; This function works like cons, but if the given element is void returns the
;; given list
(define (cons-no-void e l)
  (if (not (void? e)) (cons e l) l))




